package com.fs.starfarer.api.plugins.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.fs.starfarer.api.Global;
import com.fs.starfarer.api.campaign.CampaignFleetAPI;
import com.fs.starfarer.api.characters.MutableCharacterStatsAPI;
import com.fs.starfarer.api.characters.OfficerDataAPI;
import com.fs.starfarer.api.characters.PersonAPI;
import com.fs.starfarer.api.characters.SkillSpecAPI;
import com.fs.starfarer.api.characters.MutableCharacterStatsAPI.SkillLevelAPI;
import com.fs.starfarer.api.combat.ShipVariantAPI;
import com.fs.starfarer.api.combat.ShipAPI.HullSize;
import com.fs.starfarer.api.combat.WeaponAPI.AIHints;
import com.fs.starfarer.api.combat.WeaponAPI.WeaponType;
import com.fs.starfarer.api.fleet.FleetMemberAPI;
import com.fs.starfarer.api.impl.campaign.ids.HullMods;
import com.fs.starfarer.api.impl.campaign.ids.Tags;
import com.fs.starfarer.api.loading.FighterWingSpecAPI;
import com.fs.starfarer.api.loading.HullModSpecAPI;
import com.fs.starfarer.api.loading.WeaponGroupSpec;
import com.fs.starfarer.api.loading.WeaponSlotAPI;
import com.fs.starfarer.api.loading.WeaponSpecAPI;
import com.fs.starfarer.api.util.Misc;
import com.fs.starfarer.api.util.WeightedRandomPicker;

public class CoreAutofitPlugin extends BaseAutofitPlugin {

	public static String BUY_FROM_MARKET = new String("buy_from_market");
	public static String USE_FROM_CARGO = new String("use_from_cargo");
	public static String USE_FROM_STORAGE = new String("use_from_storage");
	public static String BUY_FROM_BLACK_MARKET = new String("black_market");
	//public static String USE_BETTER = new String("use_better");
	public static String UPGRADE = new String("upgrade");
	public static String ALWAYS_REINFORCED_HULL = new String("always_reinforced_hull");
	public static String ALWAYS_BLAST_DOORS = new String("always_blast_doors");
	public static String STRIP = new String("strip");
	//public static String USE_OTHER = new String("use_other");

	
	public static String LR = "LR";
	public static String SR = "SR";
	
	public static String KINETIC = "kinetic";
	public static String HE = "he";
	public static String ENERGY = "energy";
	public static String PD = "pd";
	public static String BEAM = "beam";
	
	public static String STRIKE = "strike";
	public static String MISSILE = "missile";
	public static String UTILITY = "utility";
	public static String ROCKET = "rocket";
	
	public static String INTERCEPTOR = "interceptor";
	public static String BOMBER = "bomber";
	public static String FIGHTER = "fighter";
	public static String SUPPORT = "support";
	
	
	public static class Category {
		public String base;
		public Set<String> tags = new HashSet<String>();
		
		public List<String> fallback = new ArrayList<String>();
		
		public Category(String base, Map<String, Category> categories) {
			this.base = base;
			categories.put(base, this);
			
			for (int i = 0; i < 100; i++) {
				String id = base + i;
				tags.add(id);
				categories.put(id, this);
			}
		}
		
		public void addFallback(String ... categories) {
			for (String catId : categories) {
				fallback.add(catId);
			}
		}
	}
	
	private List<AutofitOption> options = new ArrayList<AutofitOption>();
	
	private Map<String, Category> categories = new HashMap<String, Category>();
	
	private boolean debug = false;
	private PersonAPI fleetCommander;
	private MutableCharacterStatsAPI stats;
	
	public boolean isChecked(String id) {
		for (AutofitOption option : options) {
			if (option.id.equals(id)) return option.checked;
		}
		return false;
	}
	
	public CoreAutofitPlugin(PersonAPI fleetCommander) {
		this.fleetCommander = fleetCommander;
		if (fleetCommander != null) stats = fleetCommander.getStats();
		options.add(new AutofitOption(USE_FROM_CARGO, "Use ordnance from cargo", true, 
				"Use weapons and fighter LPCs from your fleet's cargo holds."));
		options.add(new AutofitOption(USE_FROM_STORAGE, "Use ordnance from storage", true, 
				"Use weapons and fighter LPCs from your local storage facilities."));
		options.add(new AutofitOption(BUY_FROM_MARKET, "Buy ordnance from market", true, 
				"Buy weapons and fighter LPCs from market, if docked at one.\n\n" +
				"Ordnance from your cargo will be preferred if that option is checked and if the alternatives are of equal quality."));
		options.add(new AutofitOption(BUY_FROM_BLACK_MARKET, "Allow black market purchases", true, 
				"Buy weapons and fighter LPCs from the black market.\n\n" +
				"Non-black-market options will be preferred if the alternatives are of equal quality."));
		options.add(new AutofitOption(UPGRADE, "Upgrade weapons if there are OP remaining", true, 
				"Use weapons better than the ones specified in the goal variant, if there are ordnance points left to mount them.\n\n" +
				"Will add flux vents and capacitors up to the number specified in the goal variant first, " +
				"then upgrade weapons, and then add more vents and some common hullmods.\n\n" +
				"Leaving some unspent ordnance points in a goal variant can help take advantage of this option."));
		options.add(new AutofitOption(STRIP, "Strip before autofitting", true, 
				"Remove everything possible prior to autofitting; generally results in a better fit.\n\n" +
				"However, refitting outside of port reduces a ship's combat readiness, and this option tends to lead to more changes and more readiness lost."));
		options.add(new AutofitOption(ALWAYS_REINFORCED_HULL, "Always add \"Reinforced Bulkheads\"", false, 
				"Prioritizes installing the \"Reinforced Bulkheads\" hullmod, which increases hull integrity and " +
				"makes a ship virtually certain to be recoverable if lost in battle.\n\n" +
				"\"Reinforced Bulkheads\" may still be added if this option isn't checked, provided there are enough ordnance points."));
		options.add(new AutofitOption(ALWAYS_BLAST_DOORS, "Always add \"Blast Doors\"", false, 
				"Prioritizes installing the \"Blast Doors\" hullmod, which increases hull integrity and " +
				"greatly reduces crew losses suffered due to hull damage.\n\n" +
				"\"Blast Doors\" may still be added if this option isn't checked, provided there are enough ordnance points."));
		
		
		new Category(KINETIC, categories).addFallback(KINETIC, ENERGY, HE, BEAM, PD, ROCKET, MISSILE, UTILITY, STRIKE);
		new Category(HE, categories).addFallback(HE, ENERGY, KINETIC, BEAM, PD, ROCKET, MISSILE, UTILITY, STRIKE);
		new Category(ENERGY, categories).addFallback(ENERGY, KINETIC, HE, BEAM, PD, ROCKET, MISSILE, UTILITY, STRIKE);
		new Category(PD, categories).addFallback(PD, BEAM, HE, KINETIC, UTILITY, ROCKET, MISSILE, STRIKE);
		new Category(BEAM, categories).addFallback(BEAM, ENERGY, HE, KINETIC, ROCKET, MISSILE, UTILITY, STRIKE);
		
		new Category(STRIKE, categories).addFallback(STRIKE, MISSILE, ROCKET, HE, ENERGY, KINETIC, UTILITY, BEAM, PD);
		new Category(MISSILE, categories).addFallback(MISSILE, STRIKE, ROCKET, HE, ENERGY, KINETIC, UTILITY, BEAM, PD);
		new Category(UTILITY, categories).addFallback(UTILITY, MISSILE, ROCKET, STRIKE, HE, KINETIC, ENERGY, BEAM, PD);
		new Category(ROCKET, categories).addFallback(ROCKET, UTILITY, MISSILE, STRIKE, HE, ENERGY, KINETIC, BEAM, PD);
		
		new Category(INTERCEPTOR, categories).addFallback(INTERCEPTOR, FIGHTER, SUPPORT, BOMBER);
		new Category(BOMBER, categories).addFallback(BOMBER, FIGHTER, INTERCEPTOR, SUPPORT);
		new Category(FIGHTER, categories).addFallback(FIGHTER, INTERCEPTOR, BOMBER, SUPPORT);
		new Category(SUPPORT, categories).addFallback(SUPPORT, INTERCEPTOR, FIGHTER, BOMBER);
		
		
	}
	
	
	protected void stripWeapons(ShipVariantAPI current, AutofitPluginDelegate delegate) {
		for (String id : current.getFittedWeaponSlots()) {
			WeaponSlotAPI slot = current.getSlot(id);
			if (slot.isDecorative() || slot.isBuiltIn() || slot.isHidden() ||
					slot.isSystemSlot() || slot.isStationModule()) continue;
			clearWeaponSlot(slot, delegate);
		}
	}
	
	protected void stripFighters(ShipVariantAPI current, AutofitPluginDelegate delegate) {
		int numBays = 20; // well above whatever it might actually be
		for (int i = 0; i < numBays; i++) {
			if (current.getWingId(i) != null) {
				clearFighterSlot(i, delegate);
			}
		}
	}
	
	protected Map<WeaponSlotAPI, AvailableWeapon> fittedWeapons = new HashMap<WeaponSlotAPI, AvailableWeapon>();
	protected Map<Integer, AvailableFighter> fittedFighters = new HashMap<Integer, AvailableFighter>();
	
	
	public int getCreditCost() {
		int cost = 0;
		for (AvailableWeapon w : fittedWeapons.values()) {
			cost += w.getPrice();
		}
		for (AvailableFighter w : fittedFighters.values()) {
			cost += w.getPrice();
		}
		return cost;
	}
	
	private Set<String> slotsToSkip = new HashSet<String>();
	private Set<Integer> baysToSkip = new HashSet<Integer>();
	public void doFit(ShipVariantAPI current, ShipVariantAPI target, AutofitPluginDelegate delegate) {
		fittedWeapons.clear();
		fittedFighters.clear();
		
		slotsToSkip.clear();
		baysToSkip.clear();
		
		boolean strip = isChecked(STRIP);
		if (strip) {
			stripWeapons(current, delegate);
			stripFighters(current, delegate);
			
			current.setNumFluxCapacitors(0);
			current.setNumFluxVents(0);
			current.clearHullMods();
		} else {
			slotsToSkip.addAll(current.getFittedWeaponSlots()); 
			for (int i = 0; i < 20; i++) {
				String wingId = current.getWingId(i);
				if (wingId != null && !wingId.isEmpty()) {
					baysToSkip.add(i);
				}
			}
		}
		
		List<String> mods = delegate.getAvailableHullmods();
		
		boolean reinforcedHull = isChecked(ALWAYS_REINFORCED_HULL);
		boolean blastDoors = isChecked(ALWAYS_BLAST_DOORS);
		
		if (reinforcedHull) {
			addHullmods(current, mods, HullMods.REINFORCEDHULL);
		}
		if (blastDoors) {
			addHullmods(current, mods, HullMods.BLAST_DOORS);
		}

		addHullmods(current, mods, target.getNonBuiltInHullmods().toArray(new String[0]));
		
		fitFighters(current, target, false, delegate);
		fitWeapons(current, target, false, delegate);

		if (current.hasHullMod(HullMods.FRAGILE_SUBSYSTEMS) && 
				(current.getHullSize() == HullSize.FRIGATE || current.getHullSize() == HullSize.DESTROYER)) {
			addHullmods(current, mods, HullMods.HARDENED_SUBSYSTEMS);
		}
		
		addVentsAndCaps(current, target);

		
		// now that we're at the target level of vents and caps
		// see if we can upgrade some weapons
		boolean upgrade = isChecked(UPGRADE);
		if (upgrade) {
			fitFighters(current, target, true, delegate);
			fitWeapons(current, target, true, delegate);
		}
		
//		float dissipation = current.getHullSpec().getFluxDissipation() + current.getNumFluxVents() * 10f;
//		float generation = 0f;
//		for (String slotId : current.getFittedWeaponSlots()) {
//			WeaponSpecAPI spec = current.getWeaponSpec(slotId);
//			generation += spec.getDerivedStats().getSustainedFluxPerSecond();
//		}
		
		addHullmods(current, mods, HullMods.REINFORCEDHULL, HullMods.BLAST_DOORS, HullMods.HARDENED_SUBSYSTEMS);
		addExtraVentsAndCaps(current);
		addModsWithSpareOPIfAny(current, target);
		
		current.setVariantDisplayName(target.getDisplayName());
		
		current.getWeaponGroups().clear();
		for (WeaponGroupSpec group : target.getWeaponGroups()) {
			WeaponGroupSpec copy = new WeaponGroupSpec(group.getType());
			copy.setAutofireOnByDefault(group.isAutofireOnByDefault());
			for (String slotId : group.getSlots()) {
				if (current.getWeaponId(slotId) != null) {
					copy.addSlot(slotId);
				}
			}
			if (!copy.getSlots().isEmpty()) {
				current.addWeaponGroup(copy);
			}
		}
		if (current.getWeaponGroups().isEmpty()) {
			current.autoGenerateWeaponGroups();
		}
		current.assignUnassignedWeapons();
		
		
		delegate.syncUIWithVariant();
	}
	
	private void addModsWithSpareOPIfAny(ShipVariantAPI current, ShipVariantAPI target) {
		int opCost = current.computeOPCost(stats);
		int opMax = current.getHullSpec().getOrdnancePoints(stats);
		int opLeft = opMax - opCost;
	
		if (opLeft <= 0) return;
		
		float total = target.getNumFluxVents() + target.getNumFluxCapacitors();
		float ventsFraction = 1f;
		if (total > 0) {
			ventsFraction = target.getNumFluxVents() / total;
		}
		
		if (ventsFraction >= 0.5f) {
			addDistributor(current);
			addCoil(current);
		} else {
			addCoil(current);
			addDistributor(current);
		}
	}
	
	private void addCoil(ShipVariantAPI current) {
		int opCost = current.computeOPCost(stats);
		int opMax = current.getHullSpec().getOrdnancePoints(stats);
		int opLeft = opMax - opCost;
	
		if (opLeft <= 0) return;
		
		int vents = current.getNumFluxVents();
		
		HullModSpecAPI coil = Misc.getMod(HullMods.FLUX_COIL);
		int cost = coil.getCostFor(current.getHullSize());
		
		if (cost <= opLeft + vents * 0.3f) {
			int remove = cost - opLeft;
			if (remove > 0) {
				opLeft -= addVents(-remove, current, 1000);
			}
			opLeft -= addModIfPossible(HullMods.FLUX_COIL, current, opLeft);
		}
	}
	
	private void addDistributor(ShipVariantAPI current) {
		int opCost = current.computeOPCost(stats);
		int opMax = current.getHullSpec().getOrdnancePoints(stats);
		int opLeft = opMax - opCost;
		
		if (opLeft <= 0) return;
		
		int caps = current.getNumFluxCapacitors();
		
		HullModSpecAPI distributor = Misc.getMod(HullMods.FLUX_DISTRIBUTOR);
		int cost = distributor.getCostFor(current.getHullSize());
		
		if (cost <= opLeft + caps * 0.3f) {
			int remove = cost - opLeft;
			if (remove > 0) {
				opLeft -= addCapacitors(-remove, current, 1000);
			}
			opLeft -= addModIfPossible(HullMods.FLUX_DISTRIBUTOR, current, opLeft);
		}
	}

	
	
	protected List<AvailableWeapon> getWeapons(AutofitPluginDelegate delegate) {
		boolean buy = isChecked(BUY_FROM_MARKET);
		boolean storage = isChecked(USE_FROM_STORAGE);
		boolean useCargo = isChecked(USE_FROM_CARGO);
		boolean useBlack = isChecked(BUY_FROM_BLACK_MARKET);
		
		List<AvailableWeapon> weapons = new ArrayList<AvailableWeapon>(delegate.getAvailableWeapons());
		
		Iterator<AvailableWeapon> iter = weapons.iterator();
		while (iter.hasNext()) {
			AvailableWeapon w = iter.next();
			if ((!buy && w.getPrice() > 0) ||
				(!storage && w.getPrice() <= 0 && w.getSubmarket() != null) ||
				(!useCargo && w.getSubmarket() == null) ||
				(!useBlack && w.getSubmarket() != null && w.getSubmarket().getPlugin().isBlackMarket())) {
				iter.remove();
			}
		}
		return weapons;
	}
	
	protected List<AvailableFighter> getFighters(AutofitPluginDelegate delegate) {
		boolean buy = isChecked(BUY_FROM_MARKET);
		boolean storage = isChecked(USE_FROM_STORAGE);
		boolean useCargo = isChecked(USE_FROM_CARGO);
		boolean useBlack = isChecked(BUY_FROM_BLACK_MARKET);
		
		List<AvailableFighter> fighters = new ArrayList<AvailableFighter>(delegate.getAvailableFighters());
		Iterator<AvailableFighter> iter = fighters.iterator();
		while (iter.hasNext()) {
			AvailableFighter f = iter.next();
			if ((!buy && f.getPrice() > 0) ||
				(!storage && f.getPrice() <= 0 && f.getSubmarket() != null) ||
				(!useCargo && f.getSubmarket() == null) ||
				(!useBlack && f.getSubmarket() != null && f.getSubmarket().getPlugin().isBlackMarket())) {
				iter.remove();
			}
		}
		return fighters;
	}
	
	public void addHullmods(ShipVariantAPI current, List<String> available, String ... mods) {
		int opCost = current.computeOPCost(stats);
		int opMax = current.getHullSpec().getOrdnancePoints(stats);
		int opLeft = opMax - opCost;
		
		for (String mod : mods) {
			if (!available.contains(mod)) {
				if (mod.equals(HullMods.INTEGRATED_TARGETING_UNIT) && 
						current.getHullSize().ordinal() >= HullSize.CRUISER.ordinal()) {
					mod = HullMods.DEDICATED_TARGETING_CORE;
				} else {
					continue;
				}
			}
			opLeft -= addModIfPossible(mod, current, opLeft);
		}
	}
	
	public int addModIfPossible(String id, ShipVariantAPI current, int opLeft) {
		if (current.hasHullMod(id)) return 0;
		
		HullModSpecAPI mod = Misc.getMod(id);
		if (mod == null) return 0;
		
		int cost = mod.getCostFor(current.getHullSize());
		if (cost > opLeft) return 0;
		
		current.addMod(id);
		return cost;
	}
	
	
	
	public void addVentsAndCaps(ShipVariantAPI current, ShipVariantAPI target) {
		int opCost = current.computeOPCost(stats);
		int opMax = current.getHullSpec().getOrdnancePoints(stats);
		int opLeft = opMax - opCost;
		
		int maxVents = getMaxVents(current.getHullSize());
		int maxCapacitors = getMaxCaps(current.getHullSize());
		
		int add = Math.max(target.getNumFluxVents() - current.getNumFluxVents(), 0);
		if (add > opLeft) add = opLeft;
		opLeft -= addVents(add, current, maxVents);
		
		add = Math.max(target.getNumFluxCapacitors() - current.getNumFluxCapacitors(), 0);
		if (add > opLeft) add = opLeft;
		opLeft -= addCapacitors(add, current, maxCapacitors);
	}
	
	public void addExtraVents(ShipVariantAPI current) {
		int opCost = current.computeOPCost(stats);
		int opMax = current.getHullSpec().getOrdnancePoints(stats);
		int opLeft = opMax - opCost;
		
		if (opLeft > 0) {
			int maxVents = getMaxVents(current.getHullSize());
			opLeft -= addVents((int) opLeft, current, maxVents);
		}
	}
	
	public void addExtraVentsAndCaps(ShipVariantAPI current) {
		int opCost = current.computeOPCost(stats);
		int opMax = current.getHullSpec().getOrdnancePoints(stats);
		int opLeft = opMax - opCost;
		
		if (opLeft > 0) {
			int maxVents = getMaxVents(current.getHullSize());
			int maxCapacitors = getMaxCaps(current.getHullSize());
			
			float total = current.getNumFluxVents() + current.getNumFluxCapacitors();
			float ventsFraction = 1f;
			if (total > 0) {
				ventsFraction = current.getNumFluxVents() / total;
			}
			
			int add = (int) (opLeft * ventsFraction);
			opLeft -= addVents(add, current, maxVents);
			add = opLeft;
			opLeft -= addCapacitors(add, current, maxCapacitors);
			
			add = opLeft;
			opLeft -= addVents(add, current, maxVents);
		}
	}
	
	public int getMaxVents(HullSize size) {
		int maxVents = getBaseMax(size);
		if (stats != null) {
			maxVents = (int) stats.getMaxVentsBonus().computeEffective(maxVents);
		}
		return maxVents;
	}
	
	public int getMaxCaps(HullSize size) {
		int maxCapacitors = getBaseMax(size);
		if (stats != null) {
			maxCapacitors = (int) stats.getMaxCapacitorsBonus().computeEffective(maxCapacitors);
		}
		return maxCapacitors;
	}
	
	public int getBaseMax(HullSize size) {
		int max = 100;
		switch (size) {
		case CAPITAL_SHIP: max = 50; break;
		case CRUISER: max = 30; break;
		case DESTROYER:	max = 20; break;
		case FRIGATE: max = 10; break;
		case FIGHTER: max = 5; break;
		}
		return max;
	}
	
	public int addVents(int add, ShipVariantAPI current, int max) {
		int target = current.getNumFluxVents() + add;
		if (target > max) target = max;
		if (target < 0) target = 0;
		int actual = target - current.getNumFluxVents();
		current.setNumFluxVents(target);
		return actual;
	}
	
	public int addCapacitors(int add, ShipVariantAPI current, int max) {
		int target = current.getNumFluxCapacitors() + add;
		if (target > max) target = max;
		if (target < 0) target = 0;
		int actual = target - current.getNumFluxCapacitors();
		current.setNumFluxCapacitors(target);
		return actual;
	}

	public void clearWeaponSlot(WeaponSlotAPI slot, AutofitPluginDelegate delegate) {
		fittedWeapons.remove(slot);
		delegate.clearWeaponSlot(slot);
	}
	
	public void clearFighterSlot(int index, AutofitPluginDelegate delegate) {
		fittedFighters.remove(index);
		delegate.clearFighterSlot(index);
	}

	public void fitWeapons(ShipVariantAPI current, ShipVariantAPI target, boolean upgradeMode, AutofitPluginDelegate delegate) {
		
		upgradeMode= false;
		//boolean upgradeWhenNothingMatchingInPrimary = isChecked(UPGRADE);
		
		Set<String> alreadyUsed = new HashSet<String>();
		for (WeaponSlotAPI slot : getWeaponSlotsInPriorityOrder(current, target, upgradeMode)) {
			if (slotsToSkip.contains(slot.getId())) continue;
			
			float opCost = current.computeOPCost(stats);
			float opMax = current.getHullSpec().getOrdnancePoints(stats);
			float opLeft = opMax - opCost;
			
			float levelToBeat = -1;
			if (upgradeMode) {
				WeaponSpecAPI curr = current.getWeaponSpec(slot.getId());
				if (curr != null) {
					float cost = curr.getOrdnancePointCost(stats, current.getStatsForOpCosts());
					opLeft += cost;
					
					for (String tag : curr.getTags()) {
						levelToBeat = Math.max(levelToBeat, getLevel(tag));
					}
				}
			}
			
			List<AvailableWeapon> weapons = getWeapons(delegate);
			List<AvailableWeapon> possible = getPossibleWeapons(slot, current, opLeft, weapons);
			if (possible.isEmpty()) continue;
			
			WeaponSpecAPI desired = target.getWeaponSpec(slot.getId());
			 // shouldn't happen since it should be filtered out by getWeaponSlotsInPriorityOrder()
			if (desired == null) continue;
			
			List<String> categories = getCategoriesInPriorityOrder(desired.getTags()); 
			
			
			AvailableWeapon pick = null;
			for (String catId : categories) {
				pick = getBestMatch(desired, upgradeMode, catId, alreadyUsed, possible);
				if (pick != null) {
					break;
				}
				if (upgradeMode) break; // don't pick from secondary categories when upgrading
			}
			
			if (pick == null && !upgradeMode) {
				OUTER: for (String catId : categories) {
					Category cat = this.categories.get(catId);
					if (cat == null) continue;
					
					for (String fallbackCatId : cat.fallback) {
						pick = getBestMatch(desired, true, fallbackCatId, alreadyUsed, possible);
						if (pick != null) {
							break OUTER;
						}
					}
				}
			}
			
			if (pick != null) {
				if (upgradeMode) {
					float pickLevel = -1;
					if (!categories.isEmpty()) {
						Category cat = this.categories.get(categories.get(0));
						if (cat != null) {
							String tag = getCategoryTag(cat, pick.getSpec().getTags());
							pickLevel = getLevel(tag);
						}
					}
					if (pickLevel <= levelToBeat) continue;
				}
				
				alreadyUsed.add(pick.getId());
				
				clearWeaponSlot(slot, delegate);
				delegate.fitWeaponInSlot(slot, pick);
				fittedWeapons.put(slot, pick);
			}
		}
		
	}
	
	
	public void fitFighters(ShipVariantAPI current, ShipVariantAPI target, boolean upgradeMode, AutofitPluginDelegate delegate) {
		
		int numBays = Global.getSettings().computeNumFighterBays(current);
		
		Set<String> alreadyUsed = new HashSet<String>();
		
		for (int i = 0; i < numBays; i++) {
			if (baysToSkip.contains(i)) continue;
			
			float opCost = current.computeOPCost(stats);
			float opMax = current.getHullSpec().getOrdnancePoints(stats);
			float opLeft = opMax - opCost;
			
			float levelToBeat = -1;
			if (upgradeMode) {
				FighterWingSpecAPI curr = current.getWing(i);
				if (curr != null) {
					float cost = curr.getOpCost();
					opLeft += cost;
					
					for (String tag : curr.getTags()) {
						levelToBeat = Math.max(levelToBeat, getLevel(tag));
					}
				}
			} else {
				if (current.getWingId(i) != null) {
					continue;
				}
			}
			
			List<AvailableFighter> fighters = getFighters(delegate);
			List<AvailableFighter> possible = getPossibleFighters(opLeft, fighters);
			if (possible.isEmpty()) continue;
			
			String desiredWingId = target.getWingId(i);
			if (desiredWingId == null || desiredWingId.isEmpty()) continue;
			
			FighterWingSpecAPI desired = Global.getSettings().getFighterWingSpec(desiredWingId);
			if (desired == null) continue;
			
			List<String> categories = getCategoriesInPriorityOrder(desired.getTags()); 
			
			
			AvailableFighter pick = null;
			for (String catId : categories) {
				pick = getBestMatch(desired, upgradeMode, catId, alreadyUsed, possible);
				if (pick != null) {
					break;
				}
				if (upgradeMode) break; // don't pick from secondary categories when upgrading
			}
			
			if (pick == null && !upgradeMode) {
				OUTER: for (String catId : categories) {
					Category cat = this.categories.get(catId);
					if (cat == null) continue;
					
					for (String fallbackCatId : cat.fallback) {
						pick = getBestMatch(desired, true, fallbackCatId, alreadyUsed, possible);
						if (pick != null) {
							break OUTER;
						}
					}
				}
			}
			
			if (pick != null) {
				if (upgradeMode) {
					float pickLevel = -1;
					if (!categories.isEmpty()) {
						Category cat = this.categories.get(categories.get(0));
						if (cat != null) {
							String tag = getCategoryTag(cat, pick.getWingSpec().getTags());
							pickLevel = getLevel(tag);
						}
					}
					if (pickLevel <= levelToBeat) continue;
				}
				
				alreadyUsed.add(pick.getId());
				
				clearFighterSlot(i, delegate);
				delegate.fitFighterInSlot(i, pick);
				fittedFighters.put(i, pick);
			}
		}
		
	}
	
	
	
	
	public AvailableWeapon getBestMatch(WeaponSpecAPI desired, boolean useBetter,
			   							String catId, Set<String> alreadyUsed, List<AvailableWeapon> possible) {
		//AvailableWeapon best = null;
		float bestScore = -1f;
		
		Category cat = categories.get(catId);
		if (cat == null) return null;
		
		String desiredTag = getCategoryTag(cat, desired.getTags());
		float desiredLevel = getLevel(desiredTag);
		
		if (desiredTag == null) {
			// fallback to categories that aren't in the tags of the desired weapon
//			for (String tag : desired.getTags()) {
//				desiredLevel = Math.max(desiredLevel, getLevel(tag));
//			}
			desiredLevel = 1000f;
		}
		
		boolean longRange = desired.hasTag(LR);
		boolean shortRange = desired.hasTag(SR);
		boolean midRange = !longRange && !shortRange;
		boolean desiredPD = desired.getAIHints().contains(AIHints.PD);
		
		WeightedRandomPicker<AvailableWeapon> best = new WeightedRandomPicker<AvailableWeapon>();
		
		for (AvailableWeapon w : possible) {
			WeaponSpecAPI spec = w.getSpec();
			String catTag = getCategoryTag(cat, spec.getTags());
			if (catTag == null) continue; // not in this category
			
			boolean currLongRange = spec.hasTag(LR);
			boolean currShortRange = spec.hasTag(SR);
			boolean currMidRange = !currLongRange && !currShortRange;
			
			// don't fit short-range weapons instead of long-range ones unless it's PD 
			if (!desiredPD && currShortRange && (midRange || longRange)) continue;
			//if (currMidRange && longRange) continue;
			
			float level = getLevel(catTag);
			if (!useBetter && level > desiredLevel) continue;
			
			float score = level;
			if (score > bestScore) {
				//best = w;
				best.clear();
				best.add(w);
				bestScore = score;
			} else if (score == bestScore) {
				best.add(w);
			}
		}
		
		
		// if the best-match tier includes the weapon specified in the target variant, use that
		// prefer one we already have to buying
		List<AvailableWeapon> allMatches = new ArrayList<AvailableWeapon>();
		List<AvailableWeapon> freeMatches = new ArrayList<AvailableWeapon>();
		for (AvailableWeapon w : best.getItems()) {
			if (desired.getWeaponId().equals(w.getId())) {
				allMatches.add(w);
				if (w.getPrice() <= 0) {
					freeMatches.add(w);
				}
			}
		}
		if (!freeMatches.isEmpty()) return freeMatches.get(0);
		if (!allMatches.isEmpty()) return allMatches.get(0);
		
		// if the best-match tier includes a weapon that we already own, filter out all non-free ones
		boolean hasFree = false;
		boolean hasNonBlackMarket = false;
		for (AvailableWeapon w : best.getItems()) {
			if (w.getPrice() <= 0) {
				hasFree = true;
			}
			if (w.getSubmarket() == null || !w.getSubmarket().getPlugin().isBlackMarket()) {
				hasNonBlackMarket = true;
			}
		}
		if (hasFree) {
			for (AvailableWeapon w : new ArrayList<AvailableWeapon>(best.getItems())) {
				if (w.getPrice() > 0) {
					best.remove(w);	
				}
			}
		} else if (hasNonBlackMarket) {
			for (AvailableWeapon w : new ArrayList<AvailableWeapon>(best.getItems())) {
				if (w.getSubmarket() != null && w.getSubmarket().getPlugin().isBlackMarket()) {
					best.remove(w);	
				}
			}
		}
		
		// if the best-match tier includes a weapon we used already, use that
		if (!alreadyUsed.isEmpty()) {
			for (AvailableWeapon w : best.getItems()) {
				if (alreadyUsed.contains(w.getId())) return w;
			}
		}
		
		if (best.isEmpty()) return null;
		
		//return best.getItems().get(0);
		return best.pick();
	}
	

	public AvailableFighter getBestMatch(FighterWingSpecAPI desired, boolean useBetter,
										String catId, Set<String> alreadyUsed, List<AvailableFighter> possible) {
		float bestScore = -1f;

		Category cat = categories.get(catId);
		if (cat == null) return null;

		String desiredTag = getCategoryTag(cat, desired.getTags());
		float desiredLevel = getLevel(desiredTag);

		WeightedRandomPicker<AvailableFighter> best = new WeightedRandomPicker<AvailableFighter>();

		for (AvailableFighter f : possible) {
			FighterWingSpecAPI spec = f.getWingSpec();
			String catTag = getCategoryTag(cat, spec.getTags());
			if (catTag == null) continue; // not in this category

			float level = getLevel(catTag);
			if (!useBetter && level > desiredLevel) continue;

			float score = level;
			if (score > bestScore) {
				best.clear();
				best.add(f);
				bestScore = score;
			} else if (score == bestScore) {
				best.add(f);
			}
		}


		// if the best-match tier includes the fighter specified in the target variant, use that
		List<AvailableFighter> allMatches = new ArrayList<AvailableFighter>();
		List<AvailableFighter> freeMatches = new ArrayList<AvailableFighter>();
		for (AvailableFighter f : best.getItems()) {
			if (desired.getId().equals(f.getId())) {
				allMatches.add(f);
				if (f.getPrice() <= 0) {
					freeMatches.add(f);
				}
			}
		}
		if (!freeMatches.isEmpty()) return freeMatches.get(0);
		if (!allMatches.isEmpty()) return allMatches.get(0);

		// if the best-match tier includes a fighter that we already own, filter out all non-free ones
		// prefer one we already have to buying
		boolean hasFree = false;
		boolean hasNonBlackMarket = false;
		for (AvailableFighter f : best.getItems()) {
			if (f.getPrice() <= 0) {
				hasFree = true;
			}
			if (f.getSubmarket() == null || !f.getSubmarket().getPlugin().isBlackMarket()) {
				hasNonBlackMarket = true;
			}
		}
		if (hasFree) {
			for (AvailableFighter f : new ArrayList<AvailableFighter>(best.getItems())) {
				if (f.getPrice() > 0) {
					best.remove(f);	
				}
			}
		} else if (hasNonBlackMarket) {
			for (AvailableFighter f : new ArrayList<AvailableFighter>(best.getItems())) {
				if (f.getSubmarket() != null && f.getSubmarket().getPlugin().isBlackMarket()) {
					best.remove(f);	
				}
			}
		}

		
		// if the best-match tier includes a fighter we used already, use that
		if (!alreadyUsed.isEmpty()) {
			for (AvailableFighter f : best.getItems()) {
				if (alreadyUsed.contains(f.getId())) return f;
			}
		}

		if (best.isEmpty()) return null;

		//return best.getItems().get(0);
		return best.pick();
	}
	
	public String getCategoryTag(Category cat, Set<String> tags) {
		String catTag = null;
		for (String tag : tags) {
			if (cat.tags.contains(tag)) {
				catTag = tag;
				break;
			}
		}
		return catTag;
	}
	

	public float getLevel(String tag) {
		Category cat = categories.get(tag);
		if (cat == null) return -1f;
		try {
			return Float.parseFloat(tag.replaceAll(cat.base, ""));
		} catch (Throwable t) {
			return -1f;
		}
	}
	
	public List<String> getCategoriesInPriorityOrder(Set<String> tags) {
		final Map<String, Float> levels = new HashMap<String, Float>();
		List<String> result = new ArrayList<String>();
		for (String tag : tags) {
			float level = getLevel(tag);
			if (level < 0) continue;
			levels.put(tag, level);
			result.add(tag);
		}
		
		Collections.sort(result, new Comparator<String>() {
			public int compare(String o1, String o2) {
				return (int)Math.signum(levels.get(o2) - levels.get(o1));
			}
		});
		
		return result;
	}
	
	
	public List<WeaponSlotAPI> getWeaponSlotsInPriorityOrder(ShipVariantAPI current, ShipVariantAPI target, boolean upgradeMode) {
		List<WeaponSlotAPI> result = new ArrayList<WeaponSlotAPI>();

		for (WeaponSlotAPI slot : current.getHullSpec().getAllWeaponSlotsCopy()) {
			if (slot.isBuiltIn() || slot.isDecorative()) continue;
			if (target.getWeaponId(slot.getId()) == null) continue;
			if (!upgradeMode && current.getWeaponId(slot.getId()) != null) continue;
			result.add(slot);
		}
		
		Collections.sort(result, new Comparator<WeaponSlotAPI>() {
			public int compare(WeaponSlotAPI w1, WeaponSlotAPI w2) {
				float s1 = getSlotPriorityScore(w1);
				float s2 = getSlotPriorityScore(w2);
				return (int) Math.signum(s2 - s1);
			}
		});
		
		return result;
	}

	public float getSlotPriorityScore(WeaponSlotAPI slot) {
		float score = 0;
		
		switch (slot.getSlotSize()) {
		case LARGE: score = 10000; break;
		case MEDIUM: score = 5000; break;
		case SMALL: score = 2500; break;
		}
		float angleDiff = Misc.getAngleDiff(slot.getAngle(), 0);
		boolean front = Misc.isInArc(slot.getAngle(), slot.getArc(), 0);
		if (front) {
			//score += 10f;
			score += 180f - angleDiff;
		}
		
		return score;
	}
	
	
	
	public List<AvailableWeapon> getPossibleWeapons(WeaponSlotAPI slot, ShipVariantAPI current, float opLeft, List<AvailableWeapon> weapons) {
		List<AvailableWeapon> result = new ArrayList<AvailableWeapon>();
		
		for (AvailableWeapon w : weapons) {
			if (w.getQuantity() <= 0) continue;
			
			WeaponSpecAPI spec = w.getSpec();
			float cost = spec.getOrdnancePointCost(stats, current.getStatsForOpCosts());
			if (cost > opLeft) continue;
			if (!slot.weaponFits(spec)) continue;
			
			if (spec.getType() == WeaponType.MISSILE || spec.getAIHints().contains(AIHints.STRIKE)) {
				boolean guided = spec.getAIHints().contains(AIHints.DO_NOT_AIM);
				if (!guided) {
					boolean guidedPoor = spec.getAIHints().contains(AIHints.GUIDED_POOR);
					float angleDiff = Misc.getDistanceFromArc(slot.getAngle(), slot.getArc(), 0);
					if (angleDiff > 45 || (!guidedPoor && angleDiff > 20)) continue;
				}
			}
			
			result.add(w);
		}
		
		return result;
	}
	
	public List<AvailableFighter> getPossibleFighters(float opLeft, List<AvailableFighter> fighters) {
		List<AvailableFighter> result = new ArrayList<AvailableFighter>();
		
		for (AvailableFighter f : fighters) {
			if (f.getQuantity() <= 0) continue;
			
			FighterWingSpecAPI spec = f.getWingSpec();
			float cost = spec.getOpCost();
			if (cost > opLeft) continue;
			
			result.add(f);
		}
		
		return result;
	}
	
	
	public List<AutofitOption> getOptions() {
		return options;
	}

	public float getRating(ShipVariantAPI current, ShipVariantAPI target, AutofitPluginDelegate delegate) {
		return 0;
	}

	@Override
	public void doQuickAction(ShipVariantAPI current, AutofitPluginDelegate delegate) {
		List<String> mods = delegate.getAvailableHullmods();
		addHullmods(current, mods, HullMods.REINFORCEDHULL);
		addExtraVentsAndCaps(current);
		addModsWithSpareOPIfAny(current, current);
		addHullmods(current, mods, HullMods.REINFORCEDHULL, HullMods.BLAST_DOORS, HullMods.HARDENED_SUBSYSTEMS);
		delegate.syncUIWithVariant();
	}

	@Override
	public String getQuickActionText() {
		return "Spend Free OP";
	}
	
	public String getQuickActionTooltip() {
		return "Spend any unused ordnance points on flux vents, capacitors, and essential hullmods.\n\n" +
				//"Will not make any changes to weapon loadout or changes that would reduce the combat readiness of the ship, and will not spend any credits.";
				"Will not make any changes to weapon loadout, and will not spend any credits.";
	}
	
	public boolean isQuickActionEnabled(ShipVariantAPI currentVariant) { 
		return currentVariant.getUnusedOP(stats) > 0;
	}
	
	
	@Override
	public void autoAssignOfficers(CampaignFleetAPI fleet) {
		List<FleetMemberAPI> members = new ArrayList<FleetMemberAPI>();
		for (FleetMemberAPI member : fleet.getFleetData().getMembersListCopy()) {
			if (member.isMothballed()) {
				continue;
			}
			if (!member.getCaptain().isDefault()) {
				continue;
			}
			members.add(member);
		}

		Collections.sort(members, new Comparator<FleetMemberAPI>() {
			public int compare(FleetMemberAPI o1, FleetMemberAPI o2) {
				float w1 = (float) o1.getFleetPointCost() * getVariantOPFraction(o1);
				float w2 = (float) o2.getFleetPointCost() * getVariantOPFraction(o2);
				return (int) Math.signum(w2 - w1);
			} 
		});
		
		
		List<OfficerDataAPI> officers = new ArrayList<OfficerDataAPI>();
		for (OfficerDataAPI officer : fleet.getFleetData().getOfficersCopy()) {
			boolean found = false;
			for (FleetMemberAPI member : fleet.getFleetData().getMembersListCopy()) {
				if (member.getCaptain() == officer.getPerson()) {
					found = true;
					break;
				}
			}
			if (!found) {
				officers.add(officer);
			}
		}
		
		
		List<OfficerDataAPI> normalOfficers = new ArrayList<OfficerDataAPI>(officers);
		List<OfficerDataAPI> carrierOfficers = new ArrayList<OfficerDataAPI>(officers);
		
		Collections.sort(normalOfficers, new Comparator<OfficerDataAPI>() {
			public int compare(OfficerDataAPI o1, OfficerDataAPI o2) {
				float w1 = getSkillTotal(o1, false);
				float w2 = getSkillTotal(o2, false);
				return (int) Math.signum(w2 - w1);
			}
		});
		
		Collections.sort(carrierOfficers, new Comparator<OfficerDataAPI>() {
			public int compare(OfficerDataAPI o1, OfficerDataAPI o2) {
				float w1 = getSkillTotal(o1, true);
				float w2 = getSkillTotal(o2, true);
				return (int) Math.signum(w2 - w1);
			}
		});
		
		for (FleetMemberAPI member : members) {
			OfficerDataAPI pick = null;
			if (member.isCarrier()) {
				if (!carrierOfficers.isEmpty()) pick = carrierOfficers.get(0);
				if (pick == null && !normalOfficers.isEmpty()) pick = normalOfficers.get(0);
			} else {
				if (!normalOfficers.isEmpty()) pick = normalOfficers.get(0);
				if (pick == null && !carrierOfficers.isEmpty()) pick = carrierOfficers.get(0);
			}
			
			if (pick != null) {
				carrierOfficers.remove(pick);
				normalOfficers.remove(pick);
				
				member.setCaptain(pick.getPerson());
			}
		}
	}

	
	public float getVariantOPFraction(FleetMemberAPI member) {
		float f = 1f;
		float op = member.getVariant().getHullSpec().getOrdnancePoints(stats);
		if (op > 0) {
			f = (op - member.getVariant().getUnusedOP(stats)) / op;
		}
		return f;
	}
	
	public float getSkillTotal(OfficerDataAPI officer, boolean carrier) {
		float total = 0f;
		for (SkillLevelAPI skill : officer.getPerson().getStats().getSkillsCopy()) {
			SkillSpecAPI spec = skill.getSkill();
			if (!spec.isCombatOfficerSkill()) continue;
			
			float level = skill.getLevel();
			if (level <= 0) continue;
			
			if (!carrier || spec.hasTag(Tags.SKILL_CARRIER)) {
				total += level;
			}
		}
		return total;
	}
	
	
	
}











